
package org.ululatus.sleep.bridges;

import java.io.*;
import java.net.*;
import java.util.Vector;
import java.util.Hashtable;
import java.util.StringTokenizer;

public class ExtendableClassLoader extends ClassLoader
{

	private Hashtable extending_loaders = new Hashtable();
	private ExtendingClassLoader[] loaders;
	
	public ExtendableClassLoader()
	{
		super();
	}
	
	public ExtendableClassLoader(ClassLoader parent)
	{
		super(parent);
	}
	
	protected boolean alreadyAdded(String classpath)
	{
		return extending_loaders.containsKey( classpath );
	} 

	public boolean extendClasspath(String classpath) throws InvalidClasspathException
	{
		boolean result = false;
		ExtendingClassLoader ecl = null;
		if (!alreadyAdded(getCanonicalPath( classpath )))
		{
			ecl = new ExtendingClassLoader(this,classpath);
		}
		else result = true;
		if (ecl != null)
		{
			extending_loaders.put( ecl.getClasspath(), ecl );
			loaders = getLoaders();
			result = true;
		}
		return result;
	}
	
	public void removeAllPaths()
	{
 		for ( int i=0; i<loaders.length; i++ )
 		{
 			loaders[i].closeFile();
 		}
 		extending_loaders.clear();
	}
	
	private String getCanonicalPath( String filename ) throws InvalidClasspathException
	{
		String result = null;
		try
		{
			result = new File( filename ).getCanonicalPath();
		}
		catch ( IOException ioe )
		{
			throw new InvalidClasspathException( ioe.toString() );
		}
		return result;
	}
	
	private ExtendingClassLoader[] getLoaders()
	{
		ExtendingClassLoader[] result = new ExtendingClassLoader[extending_loaders.size()];
		result = (ExtendingClassLoader[]) extending_loaders.values().toArray( result );
		return result;
	}
	
	public Class findClass(String classname)
	{
		Class c = null;

		for (int i=0;i<1;i++){	try
		{
			c = Class.forName(classname);
			//c = Class.forName(classname,true,this);
		}
		catch (ClassNotFoundException cnf) { last_error=cnf.toString();continue;} 
		catch (NoClassDefFoundError nc) { last_error=nc.toString();continue;}
		catch (LinkageError le) { last_error=le.toString();continue;}
		catch (Exception e) { last_error=e.toString();continue;} }
		if ( c != null) return c;
		for (int i=0;i<loaders.length;i++)
		{
			c = loaders[i].findClass(classname);
			last_error = loaders[i].getLastError();
			if (c!=null)
			{
				return c;
			}
		}
		return c;
	}
	String last_error = "";
	public String getLastError()
	{
//		String last_error$ = getLastError$();
		String err = "";
//		if ( !last_error$.equals("") )
//			err = "err$:"+last_error$+" ";
		err += last_error;
		last_error = "";
		return err;
	}
	
	public String getLastError$()
	{
		String err = "";
		for (int i=0;i<loaders.length;i++)
		{
			err = loaders[i].getLastError();
			if (err.length()>0)
				return err;
		}
		return err;
	}
	
	public byte[] getBytecode(String classname)//synchronized
	{
		byte[] bytes = null;
		for (int i=0;i<loaders.length;i++)
		{
			bytes = loaders[i].loadClassData(classname);
			if (bytes!=null)
			{
				return bytes;
			}
		}
		return null;
	}
	
	
	public boolean lookup(String classname)
	{
		byte[] b = getBytecode(classname);
		if (b!=null)
			return true;
		else return false;
	}
	
	public URL findResource(String resource)
	{
		URL u = null;
		for (int i=0;i<loaders.length;i++)
		{
			u = loaders[i].findResource(resource);
			if (u!=null)
				return u;
		}
		return u;
	}
	
	public ResourceHolder locateResource(String resource)
	{
		String r = null;
		ExtendingClassLoader l = null;
		for (int i=0;i<loaders.length;i++)
		{
			r = loaders[i].locateResource(resource);
			if (r!=null)
				return new ResourceHolder(l,r);
		}
		return null;
	}
	
	class ResourceHolder
	{
		ExtendingClassLoader ld;
		String rsrc;
		ResourceHolder(ExtendingClassLoader l, String r)
		{
			ld = l;
			rsrc = r;
		}
	}
	
	public byte[] loadResource(String resource)//synchronized
	{
		byte[] res = null;
		ResourceHolder holder = locateResource(resource);
		if (holder!=null)
		{
			res = holder.ld.loadResource(holder.rsrc);
		}
		return res;
	}
	
	String[] getClasspathAsArray(String classpaths)
	{
		String psep = File.pathSeparator;
		StringTokenizer tokenizer = new StringTokenizer(classpaths,psep);
		Vector tokens = new Vector();
		for (;tokenizer.hasMoreTokens();)
		{
			tokens.addElement(tokenizer.nextToken());
		}
		String[] result = new String[tokens.size()];
		tokens.copyInto(result);
		return result;
	}
	
	public String[] getClasspaths()
	{
		String[] result = new String[0];
		if ( !extending_loaders.isEmpty() )
		{
			String[] cp = (String[]) extending_loaders.keySet().toArray(result);
			return cp;
		}
		return result;
	}
	
	public String getClasspath()
	{
		String classpath = "";
		String[] patharray = getClasspaths();
		for ( int i=0 ; i<patharray.length; i++ )
		{
			classpath += patharray[i];
			if ( i != (patharray.length-1) )
				classpath += File.pathSeparator;
		}
		return classpath;
	}

 	protected void finalize() throws Throwable
 	{
 		for ( int i=0; i<loaders.length; i++ )
 		{
 			loaders[i].closeFile();
 		}
 		extending_loaders.clear();
 	}
}